from kubernetes import client, config, watch
from concurrent.futures import ThreadPoolExecutor
import time

def update_instance(ns):
    wa = watch.Watch()
    dep = client.AppsV1Api().read_namespaced_deployment(name="chat-" + ns + "-app", namespace=ns)
    replicas = dep.spec.replicas
    dep.spec.replicas = 0

    resp = client.AppsV1Api().patch_namespaced_deployment(
         name="chat-" + ns + "-app", namespace=ns, body=dep
    )

    print(f"Instance: {ns}, dep: scaled down")

    create_pre_backup(ns)
    try: 
        wait_pre_upgrade(wa, ns)
    except Exception as e:
        print(f"Instance: {ns}, pre-backup: failed, error: {e}")
        return

    upgrade_mongo(wa, ns)

    patch_rocketchat(ns)
    wait_rocketchat(wa, ns)

    dep = client.AppsV1Api().read_namespaced_deployment(name="chat-" + ns + "-app", namespace=ns)
    dep.spec.replicas = replicas

    resp = client.AppsV1Api().patch_namespaced_deployment(
         name="chat-" + ns + "-app", namespace=ns, body=dep
    )

    print(f"Instance: {ns}, dep: scale up - {replicas}")

    upgrade_apps(wa, ns)

    return f"Instance: {ns}, upgrade: done"

def upgrade_mongo(wa, ns):
    mongo = client.CustomObjectsApi().get_namespaced_custom_object(
        group="psmdb.percona.com",
        version="v1",
        name="chat-" + ns,
        namespace=ns,
        plural="perconaservermongodbs",
    )

    mongo["spec"]["allowUnsafeConfigurations"] = False
    mongo["spec"]["pmm"]["enabled"] = False
    mongo["spec"]["backup"]["restartOnFailure"] = True
    mongo["spec"]["mongod"]["redactClientLogData"] = False

    if mongo["spec"]["image"] == "percona/percona-server-mongodb:4.4.18-18":
        print(f"Instance: {ns}, mongo: already upgraded")
        return

    mongo["spec"]["image"] = "percona/percona-server-mongodb:4.4.18-18"

    patch_resource = client.CustomObjectsApi().patch_namespaced_custom_object(
        group="psmdb.percona.com",
        version="v1",
        name="chat-" + ns,
        namespace=ns,
        plural="perconaservermongodbs",
        body=mongo,
    )
    
    print(f"Instance: {ns}, mongo: patching")

    wait_mongo(wa, ns)
    create_upgrade_mongo_job(ns)
    wait_job(wa, ns, "mongo-upgrade-v4.4")

def create_pre_backup(ns):
    print(f"Instance: {ns}, prebackup: creating")
    ts = time.time()
    name = "chat-" + ns

    pre_backup = {
        "apiVersion": "psmdb.percona.com/v1",
        "kind": "PerconaServerMongoDBBackup",
        "metadata": {
            "name": "pre-migration-v5-" + str(ts), 
            "namespace": ns,
            "labels": {
              "cluster": name
            }
        },
        "spec": {
            "psmdbCluster": name,
            "storageName": "backup"
        }
    }    
    
    # create the resource
    client.CustomObjectsApi().create_namespaced_custom_object(
        group="psmdb.percona.com",
        version="v1",
        namespace=ns,
        plural="perconaservermongodbbackups",
        body=pre_backup
    )


def wait_pre_upgrade(wa, ns):
    print(f"Instance: {ns}, prepackup: waiting")
    for event in wa.stream(client.CustomObjectsApi().list_namespaced_custom_object,'psmdb.percona.com','v1',ns,'perconaservermongodbbackups'):
        obj = event["object"]
        metadata = obj.get("metadata")
        ns = metadata["namespace"]
        type = event['type']

        if type == "MODIFIED":
            status = obj.get("status")
            if status["state"] == "ready":
                wa.stop()
                return

            if status["state"] == "error":
                error = status["error"]
                raise ValueError(f"prebackup failed {error}")

def wait_mongo(wa, ns):
    print(f"Instance: {ns}, mongo: waiting")
    for event in wa.stream(client.CustomObjectsApi().list_namespaced_custom_object,'psmdb.percona.com','v1',ns,'perconaservermongodbs'):
        obj = event["object"]
        metadata = obj.get("metadata")
        
        ns = metadata["namespace"]
        type = event['type']

        if type == "MODIFIED":
            status = obj.get("status")
            if status["state"] == "ready":
                wa.stop()
                return

def wait_job(wa, ns, name):
    print(f"Instance: {ns}, job {name}: waiting")

    job = client.BatchV1Api().read_namespaced_job(name=name, namespace=ns)
    if job.status.succeeded:
        return
    if not job.status.active and job.status.failed:
        raise Exception("Job Failed")

    for event in wa.stream(
        client.BatchV1Api().list_namespaced_job,
        namespace=ns,
        label_selector=f"job-name={name}",
        timeout_seconds=0
#        _request_timeout=self.active_deadline_seconds
    ):
        o = event["object"]

        if o.status.succeeded:
            wa.stop()
            return

        if not o.status.active and o.status.failed:
            wa.stop()
            raise Exception("Job Failed")

def create_upgrade_mongo_job(ns):
    print(f"Instance: {ns}, mongo job: creating")
    container = client.V1Container(
        name="mongo",
        image="mongo",
        command=["bash", "-c", 'mongosh "mongodb+srv://$(MONGODB_CLUSTER_ADMIN_USER):$(MONGODB_CLUSTER_ADMIN_PASSWORD)@$(FQDN)-rs0.$(NS).svc.cluster.local?authSource=admin&replicaSet=rs0&w=majority&ssl=false" --eval "db.adminCommand({setFeatureCompatibilityVersion: \\"4.4\\"})"'],
        env_from=[
            client.V1EnvFromSource(
                secret_ref=client.V1SecretEnvSource(name="chat-" + ns + "-mongodb-users")
            )
        ],
        env=[
            client.V1EnvVar(
                name="FQDN",
                value="chat-" + ns,
            ),
            client.V1EnvVar(
                name="NS",
                value=ns,
            )
        ]
    )

    template = client.V1PodTemplateSpec(
        spec=client.V1PodSpec(restart_policy="Never", containers=[container]))

    spec = client.V1JobSpec(
        template=template,
        backoff_limit=4)
        
    job = client.V1Job(
        api_version="batch/v1",
        kind="Job",
        metadata=client.V1ObjectMeta(name="mongo-upgrade-v4.4"),
        spec=spec)
    
    api_response = client.BatchV1Api().create_namespaced_job(
        body=job,
        namespace=ns)

def patch_rocketchat(ns):
    print(f"Instance: {ns}, rocketchat: patching")

    dep = client.AppsV1Api().read_namespaced_deployment(name="chat-" + ns + "-app", namespace=ns)
    dep.spec.replicas = 1

    dep.spec.template.spec.containers[0].image = "libresh/rocketchat:5.4.1"
    dep.spec.template.spec.containers[0].liveness_probe = client.V1Probe(
        failure_threshold=3,
        period_seconds=10,
        success_threshold=1,
        timeout_seconds=1,
        http_get= client.V1HTTPGetAction(
            path="/health",
            port="http",
            scheme="HTTP"
        )
    )

    dep.spec.template.spec.containers[0].readiness_probe = client.V1Probe(
        failure_threshold=3,
        period_seconds=10,
        success_threshold=1,
        timeout_seconds=1,
        http_get= client.V1HTTPGetAction(
            path="/health",
            port="http",
            scheme="HTTP"
        )
    )

    for env in dep.spec.template.spec.containers[0].env:
        if env.name == "USE_NATIVE_OPLOG":
            env.name = "IGNORE_CHANGE_STREAM"

    resp = client.AppsV1Api().patch_namespaced_deployment(
        name="chat-" + ns + "-app", namespace=ns, body=dep
    )

def wait_rocketchat(wa, ns):
    print(f"Instance: {ns}, rocketchat: waiting")
    for event in wa.stream(
        client.AppsV1Api().list_namespaced_deployment,
        namespace=ns,
        label_selector=f"app.kubernetes.io/instance=chat-{ns}",
        timeout_seconds=0,
#        _request_timeout=self.active_deadline_seconds
    ):
        o = event["object"]
        s = o.status

        if (s.updated_replicas == o.spec.replicas and
                s.replicas == o.spec.replicas and
                s.available_replicas == o.spec.replicas and
                s.observed_generation >= o.metadata.generation):
            wa.stop()
            return 

def upgrade_apps(wa, ns):
    print(f"Instance: {ns}, apps job: creating")
    ts = time.time()
    name = "rc-apps-upgrade-"+ str(ts)

    container = client.V1Container(
        name="rc-apps-mgr",
        image="libresh/rc-apps-mgr-job:latest",
        env=[
            client.V1EnvVar(
                name="HOST",
                value="http://" + "chat-" + ns + "-app:3000",
            ),
            client.V1EnvVar(
                name="USERNAME",
                value_from=client.V1EnvVarSource(
                    config_map_key_ref=client.V1ConfigMapKeySelector(
                        key="ADMIN_USERNAME",
                        name="chat-" + ns + "-config"
                    )
                ),
            ),
            client.V1EnvVar(
                name="PASSWORD",
                value_from=client.V1EnvVarSource(
                    secret_key_ref=client.V1SecretKeySelector(
                        key="ADMIN_PASS",
                        name="chat-" + ns + "-app"
                    )	
                ),
            ),
        ]
    )

    # Create and configure a spec section
    template = client.V1PodTemplateSpec(
      #  metadata=client.V1ObjectMeta(labels={"app": "rocketchat"}),
        spec=client.V1PodSpec(restart_policy="Never", containers=[container]))
    # Create the specification of deployment
    spec = client.V1JobSpec(
        template=template,
        backoff_limit=4)
    
    job = client.V1Job(
        api_version="batch/v1",
        kind="Job",
        metadata=client.V1ObjectMeta(name=name),
        spec=spec)
    
    api_response = client.BatchV1Api().create_namespaced_job(
        body=job,
        namespace=ns)

    wait_job(wa, ns, name)

def main():
    # Configs can be set in Configuration class directly or using helper utility
    config.load_kube_config()
    pool = ThreadPoolExecutor(max_workers=10)

    resp = client.AppsV1Api().list_deployment_for_all_namespaces(label_selector="app.kubernetes.io/name=Rocketchat")

    instances = []
    for deployment in resp.items:
        ns = deployment.metadata.namespace
        instances.append(ns)
        
    results = pool.map(update_instance, instances)

    for res in results:
        print(res) # print results as they become available
    
    pool.shutdown()

if __name__ == '__main__':
    main()